from .query import Node, Field, Link


class Handle:
    def __init__(
        self, name=None, options=None, alias=None, items=None, mutation=False
    ):
        self.__field_name__ = name
        self.__field_options__ = options
        self.__field_alias__ = alias
        self.__node_items__ = items
        self.__mutation__ = mutation

    def __getattr__(self, name):
        assert self.__field_name__ is None, self.__field_name__
        return self.__class__(name, mutation=self.__mutation__)

    def __getitem__(self, items):
        assert self.__node_items__ is None, self.__node_items__
        if isinstance(items, Handle):
            items = [items]
        else:
            assert isinstance(items, tuple)
        self.__node_items__ = items
        return self

    def __lshift__(self, other):
        assert isinstance(other, Handle), type(other)
        assert not self.__field_options__
        assert not self.__field_alias__
        assert not self.__node_items__
        return self.__class__(
            other.__field_name__,
            other.__field_options__,
            self.__field_name__,  # alias
            other.__node_items__,
            other.__mutation__,
        )

    def __call__(self, **options):
        assert self.__field_options__ is None, self.__field_options__
        self.__field_options__ = options
        return self

    def __repr__(self) -> str:
        return "Handle[{}]({})".format(
            "Q" if not self.__mutation__ else "M", id(self)
        )


# Q for query
Q = Handle()
# M for mutation
M = Handle(mutation=True)


def build(items: list[Handle]) -> Node:
    """Builds a query or mutation

    Example:

    .. code-block:: python
        # Query
        query = build([
            Q.foo,
            Q.bar(arg=42)[
                Q.baz,
            ],
            Q.aliased << Q.field,
        ])

        # Mutation
        query = build([
            M.create(arg=42)[
                Q.date_created,
            ],
        ])

    :param items: list of fields
    :return: ready to execute query/mutation
    """
    assert (
        len({i.__mutation__ for i in items}) == 1
    ), "Only queries or mutations are allowed in the query"

    query_items = []
    is_mutation = False

    for handle in items:
        if handle.__mutation__:
            is_mutation = True

        assert handle.__field_name__ is not None
        if handle.__node_items__ is None:
            query_items.append(
                Field(
                    handle.__field_name__,
                    handle.__field_options__,
                    handle.__field_alias__,
                )
            )
        else:
            query_items.append(
                Link(  # type: ignore
                    handle.__field_name__,
                    build(handle.__node_items__),
                    handle.__field_options__,
                    handle.__field_alias__,
                )
            )
    return Node(query_items, ordered=is_mutation)



from typing import Generic, Sequence, cast
from hiku.cache import CacheSettings
from hiku.context import ExecutionContext, ExecutionContextFinal
from hiku.engine import _ExecutorType
from hiku.federation.introspection import FederatedGraphQLIntrospection
from hiku.federation.sdl import print_sdl
from hiku.graph import GraphTransformer
from hiku.graph import Union
from hiku.schema import Schema as BaseSchema
from hiku.extensions.base_extension import Extension, ExtensionsManager
from hiku.federation.graph import Graph
from hiku.federation.version import DEFAULT_FEDERATION_VERSION
from hiku.federation.utils import get_entity_types


class FederationV1EntityTransformer(GraphTransformer):
    def visit_graph(self, obj: Graph) -> Graph:  # type: ignore
        entities = get_entity_types(obj.nodes, federation_version=1)
        unions = []
        for u in obj.unions:
            if u.name == "_Entity" and entities:
                unions.append(Union("_Entity", entities))
            else:
                unions.append(u)

        return Graph(
            [self.visit(node) for node in obj.items],
            obj.data_types,
            obj.directives,
            unions,
            obj.interfaces,
            obj.enums,
            obj.scalars,
            obj.inputs,
        )


class Schema(BaseSchema, Generic[_ExecutorType]):
    """Can execute either regular or federated queries.
    Handles following fields of federated query:
        - _service
        - _entities
    """

    graph: Graph
    introspection_cls = FederatedGraphQLIntrospection
    federation_version: int

    def __init__(
        self,
        executor: _ExecutorType,
        graph: Graph,
        mutation: Graph | None = None,
        batching: bool = False,
        introspection: bool = True,
        extensions: Sequence[Extension | type[Extension]] | None = None,
        cache: CacheSettings | None = None,
        federation_version: int = DEFAULT_FEDERATION_VERSION,
    ):
        transformers: list[GraphTransformer] = []
        if federation_version == 1:
            transformers.append(FederationV1EntityTransformer())

        super().__init__(
            graph=graph,
            mutation=mutation,
            batching=batching,
            introspection=introspection,
            extensions=extensions,
            transformers=transformers,
            executor=executor,
            cache=cache,
        )
        self.federation_version = federation_version

    def _init_execution_context(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> None:
        super()._init_execution_context(execution_context, extensions_manager)
        execution_context = cast(ExecutionContextFinal, execution_context)

        if "_service" in execution_context.query.fields_map:
            execution_context.context["__sdl__"] = print_sdl(
                self.graph,
                self.mutation,
                self.federation_version,
            )



from dataclasses import dataclass
from typing import (
    Any,
    Sequence,
    cast,
    Generic,
)

from hiku.result import Proxy
from hiku.cache import CacheSettings
from hiku.context import (
    ExecutionContext,
    ExecutionContextFinal,
    create_execution_context,
)
from hiku.denormalize.graphql import DenormalizeGraphQL
from hiku.engine import _ExecutorType, Engine
from hiku.error import GraphQLError
from hiku.executors.base import (
    BaseAsyncExecutor,
    BaseSyncExecutor,
)
from hiku.extensions.base_extension import Extension, ExtensionsManager
from hiku.extensions.base_validator import QueryValidator
from hiku.graph import Graph, GraphTransformer, apply
from hiku.introspection.graphql import GraphQLIntrospection
from hiku.merge import QueryMerger
from hiku.operation import OperationType
from hiku.query import Node
from hiku.readers.graphql import parse_query, read_operation
from hiku.validate.query import validate


class ValidationError(Exception):
    def __init__(self, errors: list[str]) -> None:
        super().__init__("{} errors".format(len(errors)))
        self.errors = errors


def _run_validation(
    graph: Graph,
    query: Node,
    validators: tuple[QueryValidator, ...] | None = None,
) -> list[str]:
    errors = validate(graph, query)
    for validator in validators or ():
        errors.extend(validator.validate(query, graph))

    return errors


@dataclass
class ExecutionResult:
    data: dict[str, Any] | None
    errors: list[GraphQLError] | None
    result: Proxy | None


class Schema(Generic[_ExecutorType]):
    engine: Engine[_ExecutorType]
    graph: Graph
    mutation: Graph | None
    batching: bool
    introspection: bool

    introspection_cls = GraphQLIntrospection

    def __init__(
        self,
        executor: _ExecutorType,
        graph: Graph,
        mutation: Graph | None = None,
        batching: bool = False,
        introspection: bool = True,
        extensions: Sequence[Extension | type[Extension]] | None = None,
        transformers: list[GraphTransformer] | None = None,
        cache: CacheSettings | None = None,
    ):
        self.engine = Engine(
            executor=executor,
            cache=cache,
        )
        self.batching = batching
        self.introspection = introspection
        self.extensions = extensions or []

        if not transformers:
            transformers = []

        execution_context = create_execution_context()
        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )
        with extensions_manager.init():
            transformers.extend(execution_context.transformers)

        # introspection transformer must be the last one in chain
        if self.introspection:
            transformers.append(self.introspection_cls(graph, mutation))

        self.graph = apply(graph, transformers)
        if mutation is not None:
            self.mutation = apply(mutation, transformers)
        else:
            self.mutation = None

    def execute_sync(
        self: "Schema[BaseSyncExecutor]",
        query: str | Node,
        variables: dict[str, Any] | None = None,
        operation_name: str | None = None,
        context: dict[str, Any] | None = None,
    ) -> ExecutionResult:
        if isinstance(query, Node):
            execution_context = create_execution_context(
                query=query,
                context=context,
                query_graph=self.graph,
                mutation_graph=self.mutation,
            )
        else:
            execution_context = create_execution_context(
                query=query,
                variables=variables,
                operation_name=operation_name,
                context=context,
                query_graph=self.graph,
                mutation_graph=self.mutation,
            )

        execution_context = cast(ExecutionContextFinal, execution_context)

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )

        try:
            with extensions_manager.operation():
                self._init_execution_context(
                    execution_context, extensions_manager
                )

                with extensions_manager.execution():
                    result = self.engine.execute(execution_context)
                    execution_context.result = result

                data = DenormalizeGraphQL(
                    execution_context.graph,
                    result,
                    execution_context.operation_type_name,
                ).process(execution_context.query)

            return ExecutionResult(data, None, result)
        except ValidationError as e:
            return ExecutionResult(
                None, [GraphQLError(message) for message in e.errors], None
            )
        except GraphQLError as e:
            return ExecutionResult(None, [e], None)

    async def execute(
        self: "Schema[BaseAsyncExecutor]",
        query: str | Node,
        variables: dict[str, Any] | None = None,
        operation_name: str | None = None,
        context: dict[str, Any] | None = None,
    ) -> ExecutionResult:
        if isinstance(query, Node):
            execution_context = create_execution_context(
                query=query,
                context=context,
                query_graph=self.graph,
                mutation_graph=self.mutation,
            )
        else:
            execution_context = create_execution_context(
                query=query,
                variables=variables,
                operation_name=operation_name,
                context=context,
                query_graph=self.graph,
                mutation_graph=self.mutation,
            )

        execution_context = cast(ExecutionContextFinal, execution_context)

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )

        try:
            with extensions_manager.operation():
                self._init_execution_context(
                    execution_context, extensions_manager
                )

                with extensions_manager.execution():
                    result = await self.engine.execute(execution_context)
                    execution_context.result = result

                data = DenormalizeGraphQL(
                    execution_context.graph,
                    result,
                    execution_context.operation_type_name,
                ).process(execution_context.query)

            return ExecutionResult(data, None, result)
        except ValidationError as e:
            return ExecutionResult(
                None, [GraphQLError(message) for message in e.errors], None
            )
        except GraphQLError as e:
            return ExecutionResult(None, [e], None)

    def _validate(
        self,
        graph: Graph,
        query: Node,
        validators: tuple[QueryValidator, ...] | None = None,
    ) -> list[str]:
        return _run_validation(graph, query, validators)

    def _init_execution_context(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> None:
        with extensions_manager.parsing():
            # do not parse query if query of type Node was provided
            if (
                execution_context.graphql_document is None
                and execution_context.query is None
            ):
                assert execution_context.query_src, "query string not provided"
                execution_context.graphql_document = parse_query(
                    execution_context.query_src
                )

            if execution_context.operation is None:
                # assume that if query of type Node provided, operation
                # populated and we will not go in this branch where
                # graphql_document is None
                assert execution_context.graphql_document is not None
                try:
                    execution_context.operation = read_operation(
                        execution_context.graphql_document,
                        execution_context.variables,
                        execution_context.request_operation_name,
                    )
                except TypeError as e:
                    raise GraphQLError("Failed to read query: {}".format(e))

            execution_context.query = execution_context.operation.query

        op = execution_context.operation
        if op.type not in (OperationType.QUERY, OperationType.MUTATION):
            raise GraphQLError(
                "Unsupported operation type: {!r}".format(op.type)
            )

        # Validation works with original query, not merged one.
        # Merged query only used for execution.
        with extensions_manager.validation():
            if execution_context.errors is None:
                execution_context.errors = self._validate(
                    execution_context.graph,
                    execution_context.query,
                    execution_context.validators,
                )

            if execution_context.errors:
                raise ValidationError(errors=execution_context.errors)

        # After merge, query considered `final`, e.g. will not be changed
        # by hiku itsef and will be used as is in execution.
        merger = QueryMerger(execution_context.graph)
        execution_context.query = merger.merge(execution_context.query)
        execution_context.operation.query = execution_context.query



import contextlib
import dataclasses
import inspect
import warnings
from collections import defaultdict
from collections.abc import Hashable, Mapping, Sequence
from functools import partial
from itertools import chain, repeat
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    NoReturn,
    TypeVar,
    overload,
)

from hiku.types import OptionalMeta

from .cache import CacheInfo, CacheSettings, CacheVisitor
from .compat import ParamSpec
from .context import ExecutionContext
from .executors.base import (
    BaseAsyncExecutor,
    BaseSyncExecutor,
    SyncAsyncExecutor,
)
from .executors.queue import Queue, SubmitRes, TaskSet, Workflow
from .graph import (
    Field,
    FieldType,
    Graph,
    Link,
    LinkType,
    Many,
    Maybe,
    MaybeMany,
    Node,
    Nothing,
    One,
)
from .query import Field as QueryField
from .query import Fragment
from .query import Link as QueryLink
from .query import Node as QueryNode
from .query import QueryTransformer, QueryVisitor
from .result import ROOT, Index, Proxy, Reference
from .utils import ImmutableDict
from .utils.serialize import serialize

NodePath = tuple[str | None, ...]


def _yield_options(
    graph: Graph,
    graph_obj: Link | Field,
    query_obj: QueryField | QueryLink,
) -> Iterator[tuple[str, Any]]:
    options = query_obj.options or {}
    for option in graph_obj.options:
        value = options.get(option.name, option.default)
        optional = isinstance(option.type, OptionalMeta)
        if value is Nothing and not optional:
            raise TypeError(
                'Required option "{}" for {!r} was not provided'.format(
                    option.name, graph_obj
                )
            )
        elif option.type_info and option.type_info.type_enum is FieldType.ENUM:
            enum = graph.enums_map[option.type_info.type_name]
            yield option.name, serialize(option.type, value, enum.parse)
        elif (
            option.type_info
            and option.type_info.type_enum is FieldType.CUSTOM_SCALAR
        ):
            scalar = graph.scalars_map[option.type_info.type_name]
            yield option.name, serialize(option.type, value, scalar.parse)
        elif (
            option.type_info
            and option.type_info.type_enum is FieldType.SCALAR
            and option.type
        ):
            yield option.name, option.type.parse(value)
        elif (
            option.type_info
            and option.type_info.type_enum is FieldType.INPUT
            and option.type
        ):
            if value is Nothing and optional:
                # if value not provided for optional option, do not add it
                # to the options dict to notify value absence
                continue

            if value is None and optional:
                # if value is None for optional option, return it as None
                yield option.name, None
            else:
                input_type = graph.inputs_map[option.type_info.type_name]
                for arg in input_type.arguments:
                    if arg.name not in value and arg.default is not Nothing:
                        value[arg.name] = arg.default
                yield option.name, value
        else:
            yield option.name, value


def _get_options(
    graph: Graph,
    graph_obj: Link | Field,
    query_obj: QueryField | QueryLink,
) -> dict:
    return dict(_yield_options(graph, graph_obj, query_obj))


class InitOptions(QueryTransformer):
    def __init__(self, graph: Graph) -> None:
        self._graph = graph
        self._path = [graph.root]

    @contextlib.contextmanager
    def enter_path(self, type_: Any) -> Iterator[None]:
        try:
            if type_ is not None:
                self._path.append(type_)
            yield
        finally:
            if type_ is not None:
                self._path.pop()

    def visit_node(self, obj: QueryNode) -> QueryNode:
        fields = []
        fragments = []

        for f in obj.fields:
            if f.name == "__typename":
                fields.append(f)
            else:
                fields.append(self.visit(f))

        for fr in obj.fragments:
            with self.enter_path(self._graph.nodes_map[fr.type_name]):
                fragments.append(self.visit(fr))

        return obj.copy(fields=fields, fragments=fragments)

    def visit_field(self, obj: QueryField) -> QueryField:
        graph_obj = self._path[-1].fields_map[obj.name]
        if graph_obj.options:
            return obj.copy(options=_get_options(self._graph, graph_obj, obj))
        else:
            return obj

    def visit_link(self, obj: QueryLink) -> QueryLink:
        graph_obj = self._path[-1].fields_map[obj.name]

        if isinstance(graph_obj, Link):
            if graph_obj.type_info.type_enum is LinkType.UNION:
                self._path.append(self._graph.unions_map[graph_obj.node])
            elif graph_obj.type_info.type_enum is LinkType.INTERFACE:
                self._path.append(self._graph.interfaces_map[graph_obj.node])
            else:
                self._path.append(self._graph.nodes_map[graph_obj.node])
            try:
                node = self.visit(obj.node)
            finally:
                self._path.pop()
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            node = obj.node

        options = (
            _get_options(self._graph, graph_obj, obj)
            if graph_obj.options
            else None
        )
        return obj.copy(node=node, options=options)


@dataclasses.dataclass
class FieldInfo:
    graph_field: Field
    query_field: QueryField | QueryLink


@dataclasses.dataclass
class LinkInfo:
    graph_link: Link
    query_link: QueryLink


class SplitQuery(QueryVisitor):
    """Splits query into two groups: fields and links.
    This is needed because we execute fields and links separately.
    """

    def __init__(self, graph_node: Node) -> None:
        self._node = graph_node
        self._fields: list[tuple[Callable, FieldInfo]] = []
        self._links: list[LinkInfo] = []

    def split(
        self, query_node: QueryNode
    ) -> tuple[list[tuple[Callable, FieldInfo]], list[LinkInfo]]:
        for item in query_node.fields:
            self.visit(item)

        for fr in query_node.fragments:
            # node fragments can have different type_names
            # if node is union or inteface
            if fr.type_name == self._node.name:
                self.visit(fr)

        return self._fields, self._links

    def visit_fragment(self, obj: Fragment) -> None:
        self.visit(obj.node)

    def visit_node(self, obj: QueryNode) -> None:
        for item in obj.fields:
            self.visit(item)

    def visit_field(self, obj: QueryField) -> None:
        if obj.name == "__typename":
            return

        graph_obj = self._node.fields_map[obj.name]
        func = getattr(graph_obj.func, "__subquery__", graph_obj.func)
        self._fields.append((func, FieldInfo(graph_obj, obj)))

    def visit_link(self, obj: QueryLink) -> None:
        graph_obj = self._node.fields_map[obj.name]
        if isinstance(graph_obj, Link):
            if graph_obj.requires:
                if isinstance(graph_obj.requires, list):
                    for r in graph_obj.requires:
                        self.visit(QueryField(r))
                else:
                    self.visit(QueryField(graph_obj.requires))
            self._links.append(LinkInfo(graph_link=graph_obj, query_link=obj))
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            # `obj` here is a link, but this link is treated as a complex field
            func = getattr(graph_obj.func, "__subquery__", graph_obj.func)
            self._fields.append((func, FieldInfo(graph_obj, obj)))


class GroupQuery(QueryVisitor):
    def __init__(self, node: Node) -> None:
        self._node = node
        self._funcs: list[Callable] = []
        self._groups: list[list[FieldInfo] | LinkInfo] = []
        self._current_func = None

    def group(
        self, node: QueryNode
    ) -> list[tuple[Callable, list[FieldInfo] | LinkInfo]]:
        for item in node.fields:
            self.visit(item)
        return list(zip(self._funcs, self._groups))

    def visit_node(self, obj: QueryNode) -> NoReturn:
        raise ValueError("Unexpected value: {!r}".format(obj))

    def visit_field(self, obj: QueryField) -> None:
        graph_obj = self._node.fields_map[obj.name]
        func = getattr(graph_obj.func, "__subquery__", graph_obj.func)
        if func == self._current_func:
            assert isinstance(self._groups[-1], list)
            self._groups[-1].append(FieldInfo(graph_obj, obj))
        else:
            self._groups.append([FieldInfo(graph_obj, obj)])
            self._funcs.append(func)
            self._current_func = func

    def visit_link(self, obj: QueryLink) -> None:
        graph_obj = self._node.fields_map[obj.name]
        if graph_obj.requires:
            if isinstance(graph_obj.requires, list):
                for r in graph_obj.requires:
                    self.visit(QueryField(r))
            else:
                self.visit(QueryField(graph_obj.requires))
        self._groups.append(LinkInfo(graph_obj, obj))
        self._funcs.append(graph_obj.func)
        self._current_func = None


def _check_store_fields(
    node: Node,
    fields: list[QueryField | QueryLink],
    ids: Any | None,
    result: Any,
) -> None:
    if node.name is not None:
        assert ids is not None
        if (
            isinstance(result, Sequence)
            and len(result) == len(ids)
            and all(
                isinstance(r, Sequence) and len(r) == len(fields)
                for r in result
            )
        ):
            return
        else:
            expected = "list (len: {}) of lists (len: {})".format(
                len(ids), len(fields)
            )
    else:
        if isinstance(result, Sequence) and len(result) == len(fields):
            return
        else:
            expected = "list (len: {})".format(len(fields))
    raise TypeError(
        "Can't store field values, node: {!r}, fields: {!r}, "
        "expected: {}, returned: {!r}".format(
            node.name or "__root__", [f.name for f in fields], expected, result
        )
    )


def _is_hashable(obj: Any) -> bool:
    if not isinstance(obj, Hashable):
        return False

    try:
        hash(obj)
    except TypeError:
        return False

    return True


def update_index(
    index: Index,
    node: Node,
    ids: list,
    entries: list[dict],
) -> None:
    """Update index with data from cache"""
    for idx, entry in zip(ids, entries):
        for node_name, data in entry.items():
            if node_name == node.name:
                index[node.name][idx].update(data)
            else:
                for i, row in data.items():
                    index[node_name][i].update(row)


def store_fields(
    index: Index,
    node: Node,
    query_fields: list[QueryField | QueryLink],
    ids: Any | None,
    query_result: Any,
) -> None:
    if inspect.isgenerator(query_result):
        warnings.warn(
            "Data loading functions should not return generators",
            DeprecationWarning,
        )
        query_result = list(query_result)

    _check_store_fields(node, query_fields, ids, query_result)

    names = [f.index_key for f in query_fields]
    if node.name is not None:
        assert ids is not None
        node_idx = index[node.name]
        for i, row in zip(ids, query_result):
            node_idx[i].update(zip(names, row))
    else:
        assert ids is None
        index.root.update(zip(names, query_result))

    return None


Req = TypeVar("Req")


def link_reqs(
    index: Index, node: Node, link: Link, ids: Any
) -> list[ImmutableDict[str, Req]] | list[Req] | Req:
    """For a given link, find link `requires` values by ids."""
    if node.name is not None:
        assert ids is not None
        node_idx = index[node.name]

        if isinstance(link.requires, list):
            reqs: list[ImmutableDict[str, Req]] = []
            for i in ids:
                req: ImmutableDict = ImmutableDict(
                    (r, node_idx[i][r]) for r in link.requires
                )
                reqs.append(req)
            return reqs

        return [node_idx[i][link.requires] for i in ids]
    else:
        # TODO(mkind): add support for requires as list in root node
        return index.root[link.requires]


def link_ref_maybe(graph_link: Link, ident: Any) -> Reference | None:
    if ident is Nothing:
        return None
    else:
        if graph_link.type_info.type_enum in (
            LinkType.UNION,
            LinkType.INTERFACE,
        ):
            return Reference(ident[1].__type_name__, ident[0])
        return Reference(graph_link.node, ident)


def link_ref_one(graph_link: Link, ident: Any) -> Reference:
    assert ident is not Nothing

    if graph_link.type_info.type_enum in (LinkType.UNION, LinkType.INTERFACE):
        return Reference(ident[1].__type_name__, ident[0])
    return Reference(graph_link.node, ident)


def link_ref_many(graph_link: Link, idents: list) -> list[Reference]:
    if graph_link.type_info.type_enum in (LinkType.UNION, LinkType.INTERFACE):
        return [Reference(i[1].__type_name__, i[0]) for i in idents]
    return [Reference(graph_link.node, i) for i in idents]


def link_ref_maybe_many(
    graph_link: Link, idents: list
) -> list[Reference | None]:
    if graph_link.type_info.type_enum in (LinkType.UNION, LinkType.INTERFACE):
        return [
            Reference(i[1].__type_name__, i[0]) if i is not Nothing else None
            for i in idents
        ]
    return [
        Reference(graph_link.node, i) if i is not Nothing else None
        for i in idents
    ]


_LINK_REF_MAKER: dict[Any, Callable] = {
    Maybe: link_ref_maybe,
    One: link_ref_one,
    Many: link_ref_many,
    MaybeMany: link_ref_maybe_many,
}


HASH_HINT = "\nHint: Consider adding __hash__ method or use hashable type for '{!r}'.".format  # noqa: E501
SEQ_HINT = "\nHint: Consider using tuple instead of '{}'.".format
DATACLASS_FROZEN_HINT = "\nHint: Use @dataclass(frozen=True) on '{}'.".format
DATACLASS_FIELD_HINT = (
    "\nHint: Field '{}.{}' of type '{}' is not hashable.".format
)  # noqa: E501


def _hashable_hint(obj: Any) -> str:
    if isinstance(obj, (list, set)):
        return SEQ_HINT(type(obj).__name__)

    if isinstance(obj, Sequence) and not isinstance(obj, str):
        return _hashable_hint(obj[0])

    if dataclasses.is_dataclass(obj):
        if not getattr(obj, dataclasses._PARAMS).frozen:  # type: ignore[attr-defined]  # noqa: E501
            return DATACLASS_FROZEN_HINT(obj.__class__.__name__)

        for f in dataclasses.fields(obj):
            val = getattr(obj, f.name)
            if not _is_hashable(val):
                return DATACLASS_FIELD_HINT(
                    obj.__class__.__name__, f.name, type(val).__name__
                )

    return HASH_HINT(obj)


def _check_store_links(node: Node, link: Link, ids: Any, result: Any) -> None:
    hint = ""
    if node.name is not None and link.requires is not None:
        assert ids is not None
        if link.type_enum is Maybe or link.type_enum is One:
            if isinstance(result, Sequence) and len(result) == len(ids):
                if all(map(_is_hashable, result)):
                    return
                expected = "list of hashable objects"
                hint = _hashable_hint(result)
            else:
                expected = "list (len: {})".format(len(ids))
        elif link.type_enum is Many or link.type_enum is MaybeMany:
            if (
                isinstance(result, Sequence)
                and len(result) == len(ids)
                and all(isinstance(r, Sequence) for r in result)
            ):
                unhashable = False
                for items in result:
                    if not all(map(_is_hashable, items)):
                        unhashable = True
                        break

                if not unhashable:
                    return
                expected = "list (len: {}) of lists of hashable objects".format(
                    len(ids)
                )  # noqa: E501
                hint = _hashable_hint(result)
            else:
                expected = "list (len: {}) of lists".format(len(ids))
        else:
            raise TypeError(link.type_enum)
    else:
        if link.type_enum is Maybe or link.type_enum is One:
            if _is_hashable(result):
                return
            expected = "hashable object"
            hint = _hashable_hint(result)
        elif link.type_enum is Many:
            if isinstance(result, Sequence):
                if all(map(_is_hashable, result)):
                    return
                expected = "list of hashable objects"
                hint = _hashable_hint(result)
            else:
                expected = "list"
        elif link.type_enum is MaybeMany:
            if isinstance(result, Sequence):
                if all(map(_is_hashable, result)):
                    return
                expected = "list of hashable objects and Nothing"
                hint = _hashable_hint(result)
            else:
                expected = "list"
        else:
            raise TypeError(link.type_enum)
    raise TypeError(
        "Can't store link values, node: {!r}, link: {!r}, "
        "expected: {}, returned: {!r}{}".format(
            node.name or "__root__", link.name, expected, result, hint
        )
    )


def store_links(
    index: Index,
    node: Node,
    graph_link: Link,
    query_link: QueryLink,
    ids: Any,
    query_result: Any,
) -> None:
    _check_store_links(node, graph_link, ids, query_result)

    field_val: Callable = _LINK_REF_MAKER[graph_link.type_enum]
    if node.name is not None:
        assert ids is not None
        if graph_link.requires is None:
            query_result = repeat(query_result, len(ids))

        node_idx = index[node.name]
        for i, res in zip(ids, query_result):
            node_idx[i][query_link.index_key] = field_val(graph_link, res)
    else:
        index.root[query_link.index_key] = field_val(graph_link, query_result)

    return None


def link_result_to_ids(
    from_list: bool, link_type: Any, result: Any
) -> list:  # Const
    if from_list:
        if link_type is Maybe:
            return [i for i in result if i is not Nothing]
        elif link_type is One:
            if any(i is Nothing for i in result):
                raise TypeError(
                    "Non-optional link should not return Nothing: "
                    "{!r}".format(result)
                )
            return result
        elif link_type is Many or link_type is MaybeMany:
            return list(chain.from_iterable(result))
    else:
        if link_type is Maybe:
            return [] if result is Nothing else [result]
        elif link_type is One:
            if result is Nothing:
                raise TypeError("Non-optional link should not return Nothing")
            return [result]
        elif link_type is Many or link_type is MaybeMany:
            return result
    raise TypeError(repr([from_list, link_type]))


Dep = SubmitRes | TaskSet


class Query(Workflow):
    def __init__(
        self,
        queue: Queue,
        task_set: TaskSet,
        graph: Graph,
        query: QueryNode,
        ctx: "Context",
        cache: CacheInfo | None = None,
    ) -> None:
        self._queue = queue
        self._task_set = task_set
        self._graph = graph
        self._query = query
        self._ctx = ctx
        self._index = Index()
        self._cache = cache
        self._in_progress: defaultdict[NodePath, int] = defaultdict(int)
        self._done_callbacks: defaultdict[NodePath, list[Callable]] = (
            defaultdict(list)
        )  # noqa: E501
        self._path_callback: dict[NodePath, Callable] = {}

    def _track(self, path: NodePath) -> None:
        self._in_progress[path] += 1

    def _add_done_callback(self, path: NodePath, callback: Callable) -> None:
        self._done_callbacks[path].append(callback)

    def _untrack(self, path: NodePath) -> None:
        assert self._in_progress[path] > 0, f"Path {path} is already done"
        self._in_progress[path] -= 1
        if self._is_done(path):
            for callback in self._done_callbacks[path]:
                callback()

            parent_path = path[:-1]
            if parent_path in self._path_callback:
                self._path_callback[parent_path]()

    def _is_done(self, path: NodePath) -> bool:
        return self._in_progress[path] == 0

    def _submit(self, func: Callable, *args: Any, **kwargs: Any) -> SubmitRes:
        if _do_pass_context(func):
            return self._task_set.submit(func, self._ctx, *args, **kwargs)
        else:
            return self._task_set.submit(func, *args, **kwargs)

    def start(self) -> None:
        self.process_node(tuple(), self._graph.root, self._query, None)

    def result(self) -> Proxy:
        self._index.finish()
        return Proxy(self._index, ROOT, self._query)

    def _process_node_ordered(
        self, path: NodePath, node: Node, query: QueryNode, ids: Any
    ) -> None:
        proc_steps = GroupQuery(node).group(query)

        # recursively and sequentially schedule fields and links
        def proc(
            steps: list[tuple[Callable, list[FieldInfo] | LinkInfo]]
        ) -> None:
            step_func, step_item = steps.pop(0)
            if isinstance(step_item, list):
                self._track(path)
                dep = self._schedule_fields(
                    path, node, step_func, step_item, ids
                )
            else:
                self._track(path)
                dep = self._schedule_link(
                    path, node, step_item.graph_link, step_item.query_link, ids
                )

            if steps:
                self._queue.add_callback(dep, lambda: proc(steps))

        if proc_steps:
            proc(proc_steps)

    def process_node(
        self,
        path: NodePath,
        node: Node,
        query: QueryNode,
        ids: Any,
    ) -> None:
        path = path + (node.name,)
        self._path_callback[path] = lambda: self._untrack(path)

        if query.ordered:
            self._process_node_ordered(path, node, query, ids)
            return

        fields, links = SplitQuery(node).split(query)

        to_func: dict[str, Callable] = {}
        from_func: defaultdict[Callable, list[FieldInfo]] = defaultdict(list)
        for func, field_info in fields:
            to_func[field_info.graph_field.name] = func
            from_func[func].append(field_info)

        to_dep: dict[Callable, Dep] = {}
        for func, func_fields_info in from_func.items():
            self._track(path)
            to_dep[func] = self._schedule_fields(
                path, node, func, func_fields_info, ids
            )

        # schedule link resolve
        for link_info in links:
            graph_link = link_info.graph_link
            link = link_info.query_link

            self._track(path)
            schedule = partial(
                self._schedule_link,
                path,
                node,
                graph_link,
                link,
                ids,
            )
            if graph_link.requires:
                if isinstance(graph_link.requires, list):
                    done_link_deps: set = set()

                    def add_done_dep_callback(
                        done_deps: set,
                        dep: Dep,
                        req: Any,
                        graph_link: Link,
                        schedule: Callable,
                    ) -> None:
                        def done_cb() -> None:
                            done_deps.add(req)
                            if done_deps == set(graph_link.requires):
                                schedule()

                        self._queue.add_callback(dep, done_cb)

                    for req in graph_link.requires:
                        add_done_dep_callback(
                            done_link_deps,
                            to_dep[to_func[req]],
                            req,
                            graph_link,
                            schedule,
                        )
                else:
                    dep = to_dep[to_func[graph_link.requires]]
                    self._queue.add_callback(dep, schedule)
            else:
                schedule()

    def process_link(
        self,
        path: NodePath,
        node: Node,
        graph_link: Link,
        query_link: QueryLink,
        ids: Any,
        result: list,
    ) -> None:
        """Store Link.func result in index and Call `process_node` to schedule
        Link's fields and links"""
        if inspect.isgenerator(result):
            warnings.warn(
                "Data loading functions should not return generators",
                DeprecationWarning,
            )
            result = list(result)
        store_links(self._index, node, graph_link, query_link, ids, result)
        from_list = ids is not None and graph_link.requires is not None
        to_ids = link_result_to_ids(from_list, graph_link.type_enum, result)
        if to_ids:
            if graph_link.type_info.type_enum in (
                LinkType.UNION,
                LinkType.INTERFACE,
            ) and isinstance(to_ids, list):
                grouped_ids = defaultdict(list)
                for id_, type_ref in to_ids:
                    grouped_ids[type_ref.__type_name__].append(id_)

                for type_name, type_ids in grouped_ids.items():
                    self.process_node(
                        path,
                        self._graph.nodes_map[type_name],
                        query_link.node,
                        list(type_ids),
                    )

                # FIXME: call track len(ids) - 1 times because first track was
                #  already called by process_node for this link
                for _ in range(len(grouped_ids) - 1):
                    self._track(path)
            else:
                if graph_link.type_enum is MaybeMany:
                    to_ids = [id_ for id_ in to_ids if id_ is not Nothing]

                self.process_node(
                    path,
                    self._graph.nodes_map[graph_link.node],
                    query_link.node,
                    # TODO: you can not pass [1, Nothing] as ids
                    to_ids,
                )
        else:
            self._untrack(path)

        return None

    def _schedule_fields(
        self,
        path: NodePath,
        node: Node,
        func: Callable,
        fields_info: list[FieldInfo],
        ids: Any | None,
    ) -> SubmitRes | TaskSet:
        query_fields = [f.query_field for f in fields_info]

        dep: TaskSet | SubmitRes
        if hasattr(func, "__subquery__"):
            assert ids is not None
            dep = self._queue.fork(self._task_set)
            fields = [(f.graph_field, f.query_field) for f in fields_info]
            proc = func(fields, ids, self._queue, self._ctx, dep)
        else:
            if ids is None:
                dep = self._submit(func, query_fields)
            else:
                dep = self._submit(func, query_fields, ids)
            proc = dep.result

        def callback() -> None:
            store_fields(
                self._index,
                node,
                query_fields,
                ids,
                proc(),
            )
            self._untrack(path)

        self._queue.add_callback(dep, callback)
        return dep

    def _update_index_from_cache(
        self,
        path: NodePath,
        node: Node,
        graph_link: Link,
        query_link: QueryLink,
        ids: list[Any],
        reqs: list[Any],
    ) -> SubmitRes:
        assert self._cache is not None
        key_info = []
        for i, req in zip(ids, reqs):
            key_info.append(
                (self._cache.query_hash(self._ctx, query_link, req), i, req)
            )

        keys = set(info[0] for info in key_info)
        dep = self._submit(
            self._cache.get_many, list(keys), node.name, graph_link.name
        )

        def callback() -> None:
            result = dep.result()
            cached_data = []
            cached_ids = []
            for key, i, req in key_info:
                if key in result:
                    cached_ids.append(i)
                    cached_data.append(result[key])

            nonlocal ids
            if cached_data:
                update_index(self._index, node, cached_ids, cached_data)
                ids = [i for i in ids if i not in cached_ids]

            if ids:
                self._schedule_link(
                    path,
                    node,
                    graph_link,
                    query_link,
                    ids,
                    skip_cache=True,
                )

        self._queue.add_callback(dep, callback)
        return dep

    def _schedule_link(
        self,
        path: NodePath,
        node: Node,
        graph_link: Link,
        query_link: QueryLink,
        ids: Any,
        skip_cache: bool = False,
    ) -> SubmitRes:
        """Schedules link (submits Link.func to executor).

        If link has `requires` specified, take required values list from index
        and pass it as first argument to link.func.

        If link has `options`, pass as a first argument if `requires` not
        provided or as second if `requires` provided.

        When Link.func is executed by executor, a `process_link`
        method called with result.
        """
        args = []
        if graph_link.requires:
            # collect data for link requires from store
            reqs: Any = link_reqs(self._index, node, graph_link, ids)

            if (
                "cached" in query_link.directives_map
                and self._cache
                and not skip_cache
            ):
                return self._update_index_from_cache(
                    path, node, graph_link, query_link, ids, reqs
                )

            args.append(reqs)

        if graph_link.options:
            args.append(query_link.options)

        dep = self._submit(graph_link.func, *args)

        def callback() -> None:
            return self.process_link(
                path,
                node,
                graph_link,
                query_link,
                ids,
                dep.result(),
            )

        self._queue.add_callback(dep, callback)

        def store_link_cache() -> None:
            assert self._cache is not None
            cached = query_link.directives_map["cached"]
            reqs: Any = link_reqs(self._index, node, graph_link, ids)
            to_cache = CacheVisitor(
                self._cache, self._index, self._graph, node
            ).process(query_link, ids, reqs, self._ctx)

            self._submit(self._cache.set_many, to_cache, cached.ttl)

        if "cached" in query_link.directives_map and self._cache:
            self._add_done_callback(path + (graph_link.node,), store_link_cache)

        return dep


R = TypeVar("R")
P = ParamSpec("P")


def pass_context(func: Callable[P, R]) -> Callable[P, R]:
    """Decorator to pass context to a function as a first argument.

    Can be used on functions for ``Field`` and ``Link``.
    Can not be used on functions with ``@define`` decorator
    """
    func.__pass_context__ = True  # type: ignore[attr-defined]
    return func


def _do_pass_context(func: Callable) -> bool:
    return getattr(func, "__pass_context__", False)


class Context(Mapping):
    def __init__(self, mapping: Mapping) -> None:
        self.__mapping = mapping

    def __len__(self) -> int:
        return len(self.__mapping)

    def __iter__(self) -> Iterator:
        return iter(self.__mapping)

    def __getitem__(self, item: Any) -> Any:
        try:
            return self.__mapping[item]
        except KeyError:
            raise KeyError(
                "Key {!r} is not specified " "in the query context".format(item)
            )


# Covariant must be used because we want to accept subclasses of Executor
_ExecutorType = TypeVar(
    "_ExecutorType", covariant=True, bound=SyncAsyncExecutor
)


class Engine(Generic[_ExecutorType]):
    executor: _ExecutorType

    def __init__(
        self,
        executor: _ExecutorType,
        cache: CacheSettings | None = None,
    ) -> None:
        self.executor = executor
        self.cache_settings = cache

    def _prepare_workflow(
        self, execution_context: ExecutionContext
    ) -> tuple[Queue, Query]:
        graph = execution_context.graph
        query = execution_context.query
        ctx = execution_context.context
        operation_name = execution_context.operation_name

        query = InitOptions(graph).visit(query)
        assert query is not None
        queue = Queue(self.executor)
        task_set = queue.fork(None)
        cache = (
            CacheInfo(
                self.cache_settings,
                operation_name,
            )
            if self.cache_settings
            else None
        )
        query_workflow = Query(
            queue, task_set, graph, query, Context(ctx), cache
        )
        query_workflow.start()
        return queue, query_workflow

    @overload
    async def execute(
        self: "Engine[BaseAsyncExecutor]",
        execution_context: ExecutionContext,
    ) -> Proxy: ...

    @overload
    def execute(
        self: "Engine[BaseSyncExecutor]",
        execution_context: ExecutionContext,
    ) -> Proxy: ...

    def execute(
        self,
        execution_context: ExecutionContext,
    ) -> Any:
        queue, workflow = self._prepare_workflow(execution_context)
        return self.executor.process(queue, workflow)



from __future__ import annotations

import contextlib
import inspect
from asyncio import iscoroutinefunction
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Callable,
    Iterator,
    NamedTuple,
    Sequence,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from hiku.context import ExecutionContext


T = TypeVar("T")

AwaitableOrValue = Union[Awaitable[T], T]
AsyncIteratorOrIterator = AsyncIterator[T] | Iterator[T]
Hook = Callable[
    ["Extension", "ExecutionContext"], AsyncIteratorOrIterator[None]
]


class Extension:
    """Extension class for hooking into the GraphQL execution lifecycle.

    Extensions allow you to intercept and modify the execution flow at various
    stages of query processing. Each hook is called before and after its
    respective stage, providing opportunities for logging, monitoring,
    validation, and more.

    **Extension Lifecycle Diagram:**

    ```
    Schema Creation
    ┌─────────────────────────────────────────────────────────────┐
    │ on_init() - Called once during schema creation              │
    │ • Add transformers to execution context                     │
    │ • Setup global state, monitoring, etc.                      │
    └─────────────────────────────────────────────────────────────┘

    Query Execution (per request)
    ┌─────────────────────────────────────────────────────────────┐
    │ on_operation() - Wraps entire operation lifecycle           │
    │ ┌─────────────────────────────────────────────────────────┐ │
    │ │ on_parse() - Query parsing and AST transformation     │ │ |
    │ │ • Parse GraphQL string to AST                         │ │ |
    │ │ • Transform AST to Hiku query nodes                   │ │ |
    │ └─────────────────────────────────────────────────────────┘ │
    │ ┌─────────────────────────────────────────────────────────┐ │
    │ │ on_validate() - Query validation                      │ │ |
    │ │ • Validate query against schema                       │ │ |
    │ │ • Check field existence, types, etc.                  │ │ |
    │ │ • Set execution_context.errors if validation fails    │ │ |
    │ └─────────────────────────────────────────────────────────┘ │
    │ ┌─────────────────────────────────────────────────────────┐ │
    │ │ on_execute() - Query execution                        │ │ |
    │ │ • Execute query against data sources                  │ │ |
    │ │ • Resolve fields and relationships                    │ │ |
    │ │ • Set execution_context.result                        │ │ |
    │ │ • Denormalize result                                  │ │ |
    │ └─────────────────────────────────────────────────────────┘ │
    └─────────────────────────────────────────────────────────────┘

    **Hook execution order:**
    1. on_init() - Schema creation (once)
    2. on_operation() - Start of operation
    3.   on_parse() - Query parsing
    4.   on_validate() - Query validation
    5.   on_execute() - Query execution
    6. on_operation() - End of operation

    **ExecutionContext fields availability:**
    - on_init: Basic execution context
    - on_operation:
        before yield: query_src, variables, operation_name, context
        after yield: all fields from execution_context
    - on_parse:
        before yield: all from previous hooks
        after yield: graphql_document (AST), operation, query
    - on_validate:
        before yield: all from previous hooks
        after yield: errors
    - on_execute:
        before yield: all from previous hooks
        after yield: result

    **Hook implementation:**
    Each hook should be implemented as a generator function that yields once:
    ```python
    def on_parse(self, execution_context: ExecutionContext) -> Iterator[None]:
        # Pre-parse logic
        yield  # This is where parsing happens
        # Post-parse logic
    ```
    """

    def on_init(  # type: ignore[return]
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called once during schema creation.

        It is a good place to add transformers to the execution context.
        """
        yield None

    def on_operation(  # type: ignore[return]
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the operation step.

        A step happens after `on_init` hook and is active during the next hooks:
        - on_parse
        - on_validate
        - on_execute

        `on_operation` hook is the last hook of entrire operation lifecycle.

        Operation step is a step where the query is executed by schema.

        At this step the:
        - execution_context.query_src (if type str) is set to the query string
        - execution_context.query (if type Node) is set to the query Node
        - execution_context.variables is set to the query variables
        - execution_context.operation_name is set to the query operation name
        - execution_context.query_graph is set to the query graph
        - execution_context.mutation_graph is set to the mutation graph
        - execution_context.context is set to the context from execute argument
        """
        yield None

    def on_parse(  # type: ignore[return]
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the parsing step.

        A step happens inside `on_operation` hook and before `on_validate` hook.

        Parse step is when query string is:
        - parsed into graphql ast and will be assigned to the
          execution_context.graphql_document.
          Note that execution_context.query_src parsed only
          if execution_context.query and execution_context.graphql_document are
          empty.
        - graphql ast is transformed into
          hiku's query ast and Operation type is created and assigned to the
          execution_context.operation.
        """
        yield None

    def on_validate(  # type: ignore[return]
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the validation step.

        A step happens inside `on_operation` hook and before `on_execute` hook.

        Validation step is when hiku query is validated.
        After validation is done, if there are errors, they will be assigned
        to the execution_context.errors.
        If there are already execution_context.errors, then validation will be
        skipped (this can happen if errors attribute was set in any hook before
        or in `on_validate` hook before yield.).
        """
        yield None

    def on_execute(  # type: ignore[return]
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the execution step.

        A step happens inside `on_operation` hook and after `on_validate` hook
        and is the last hook inside `on_operation` hook.

        Execution step is when hiku query is executed by hiku engine.

        After execution, normalized result(Proxy) will be assigned
        to execution_context.result."""
        yield None


class ExtensionsManager:
    """ExtensionManager is a per/dispatch extensions manager.

    It is used to call excensions hooks in the right order.
    """

    def __init__(
        self,
        execution_context: ExecutionContext,
        extensions: Sequence[type[Extension] | Extension],
    ):
        self.execution_context = execution_context

        if not extensions:
            extensions = []

        init_extensions: list[Extension] = []

        for extension in extensions:
            if isinstance(extension, Extension):
                init_extensions.append(extension)
            else:
                init_extensions.append(extension())

        self.extensions = init_extensions

    def init(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_init.__name__, self.extensions, self.execution_context
        )

    def operation(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_operation.__name__,
            self.extensions,
            self.execution_context,
        )

    def parsing(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_parse.__name__, self.extensions, self.execution_context
        )

    def validation(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_validate.__name__,
            self.extensions,
            self.execution_context,
        )

    def execution(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_execute.__name__,
            self.extensions,
            self.execution_context,
        )


class WrappedHook(NamedTuple):
    extension: Extension
    initialized_hook: AsyncIterator[None] | Iterator[None]
    is_async: bool


class ExtensionContextManagerBase:
    __slots__ = ("hook_name", "hooks", "deprecation_message", "default_hook")

    def __init__(
        self,
        hook_name: str,
        extensions: list[Extension],
        execution_context: ExecutionContext,
    ):
        self.hook_name = hook_name
        self.hooks: list[WrappedHook] = []
        self.default_hook: Hook = getattr(Extension, self.hook_name)
        for extension in extensions:
            hook = self.get_hook(extension, execution_context)
            if hook:
                self.hooks.append(hook)

    def get_hook(
        self, extension: Extension, execution_context: ExecutionContext
    ) -> WrappedHook | None:
        hook_fn: Hook | None = getattr(type(extension), self.hook_name)
        hook_fn = hook_fn if hook_fn is not self.default_hook else None

        if hook_fn:
            if inspect.isgeneratorfunction(hook_fn):
                return WrappedHook(
                    extension, hook_fn(extension, execution_context), False
                )

            if inspect.isasyncgenfunction(hook_fn):
                return WrappedHook(
                    extension, hook_fn(extension, execution_context), True
                )

            if callable(hook_fn):
                return self.from_callable(extension, hook_fn, execution_context)

            raise ValueError(
                f"Hook {self.hook_name} on {extension} "
                f"must be callable, received {hook_fn!r}"
            )

        return None

    @staticmethod
    def from_callable(
        extension: Extension,
        func: Hook,
        execution_context: ExecutionContext,
    ) -> WrappedHook:
        if iscoroutinefunction(func):

            async def async_iterator():  # type: ignore[no-untyped-def]
                await func(extension, execution_context)
                yield

            hook = async_iterator()
            return WrappedHook(extension, hook, True)
        else:

            def iterator():  # type: ignore[no-untyped-def]
                func(extension, execution_context)
                yield

            hook = iterator()
            return WrappedHook(extension, hook, False)

    def run_hooks_sync(self, is_exit: bool = False) -> None:
        """Run extensions synchronously."""
        ctx = (
            contextlib.suppress(StopIteration, StopAsyncIteration)
            if is_exit
            else contextlib.nullcontext()
        )
        for hook in self.hooks:
            with ctx:
                if hook.is_async:
                    raise RuntimeError(
                        f"Extension hook {hook.extension}.{self.hook_name} "
                        "is async."
                    )
                else:
                    hook.initialized_hook.__next__()  # type: ignore[union-attr]

    def __enter__(self):  # type: ignore[no-untyped-def]
        self.run_hooks_sync()

    def __exit__(  # type: ignore[no-untyped-def]
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        self.run_hooks_sync(is_exit=True)

    # TODO: support async



from typing import Any

from graphql.language import ast

from hiku.query import Fragment

from ..query import (
    QueryVisitor,
    Field,
    Node,
    Link,
)


def _name(value: Any) -> ast.NameNode | None:
    return ast.NameNode(value=value) if value is not None else None


def _encode(value: Any) -> ast.ValueNode:
    if value is None:
        return ast.NullValueNode()
    elif isinstance(value, bool):
        return ast.BooleanValueNode(value=value)
    elif isinstance(value, int):
        return ast.IntValueNode(value=str(value))
    elif isinstance(value, float):
        return ast.FloatValueNode(value=str(value))
    elif isinstance(value, str):
        return ast.StringValueNode(value=value)
    elif isinstance(value, list):
        return ast.ListValueNode(values=[_encode(v) for v in value])
    elif isinstance(value, dict):
        return ast.ObjectValueNode(
            fields=[
                ast.ObjectFieldNode(name=_name(key), value=_encode(val))
                for key, val in value.items()
            ]
        )
    else:
        raise TypeError("Unsupported type: {!r}".format(value))


class Exporter(QueryVisitor):
    def visit_field(self, obj: Field) -> ast.FieldNode:
        arguments = None
        if obj.options:
            arguments = [
                ast.ArgumentNode(name=_name(key), value=_encode(val))
                for key, val in obj.options.items()
            ]
        return ast.FieldNode(
            name=_name(obj.name),
            alias=_name(obj.alias),
            arguments=arguments,
        )

    def visit_link(self, obj: Link) -> ast.FieldNode:
        arguments = None
        if obj.options:
            arguments = [
                ast.ArgumentNode(name=_name(key), value=_encode(val))
                for key, val in obj.options.items()
            ]
        return ast.FieldNode(
            name=_name(obj.name),
            alias=_name(obj.alias),
            arguments=arguments,
            selection_set=self.visit(obj.node),
        )

    def visit_fragment(self, obj: Fragment) -> Any:
        if obj.name is None:
            return ast.InlineFragmentNode(
                type_condition=(
                    ast.NamedTypeNode(
                        name=_name(obj.type_name),
                    )
                    if obj.type_name is not None
                    else None
                ),
                selection_set=self.visit(obj.node),
            )

        return ast.FragmentSpreadNode(name=_name(obj.name))

    def visit_node(self, obj: Node) -> ast.SelectionSetNode:
        selections = []
        for f in obj.fields:
            selections.append(self.visit(f))

        for fr in obj.fragments:
            selections.append(self.visit(fr))

        return ast.SelectionSetNode(selections=selections)


def export(query: Node) -> ast.DocumentNode:
    return ast.DocumentNode(
        definitions=[
            ast.OperationDefinitionNode(
                operation=ast.OperationType.QUERY,
                selection_set=Exporter().visit(query),
            )
        ]
    )



import inspect
from asyncio import (
    FIRST_COMPLETED,
    CancelledError,
    Task,
    gather,
    get_running_loop,
    wait,
)
from typing import TYPE_CHECKING, Any, Callable, Coroutine, cast

from hiku.executors.base import BaseAsyncExecutor
from hiku.result import Proxy

if TYPE_CHECKING:
    from hiku.executors.queue import Queue, Workflow


class AsyncIOExecutor(BaseAsyncExecutor):
    """AsyncIOExecutor is an executor that uses asyncio event loop to run tasks.

    By default it allows to run both synchronous and asynchronous tasks.
    To deny synchronous tasks set deny_sync to True.

    :param deny_sync: deny synchronous tasks -
                      raise TypeError if a task is not awaitable
    """

    def __init__(self, deny_sync: bool = False) -> None:
        self.deny_sync = deny_sync

    async def _wrapper(self, result: Any) -> Any:
        return result

    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> Task:
        loop = get_running_loop()

        coro = fn(*args, **kwargs)
        if not inspect.isawaitable(coro):
            if self.deny_sync:
                raise TypeError(
                    "{!r} returned non-awaitable object {!r}".format(fn, coro)
                )

            return loop.create_task(self._wrapper(coro))

        coro = cast(Coroutine, coro)
        return loop.create_task(coro)

    async def process(self, queue: "Queue", workflow: "Workflow") -> Proxy:
        try:
            while queue.__futures__:
                done, _ = await wait(  # type: ignore
                    queue.__futures__, return_when=FIRST_COMPLETED
                )
                queue.progress(done)
            return workflow.result()
        except CancelledError:
            for task in queue.__futures__:
                task.cancel()  # type: ignore
            await gather(*queue.__futures__)  # type: ignore
            raise
